""" file:  test_borehole_collection.py
    author: Jess Robertson
            Minerals Down Under
    date:   today

    description: Tests for BoreholeCollection class.
"""

from pysiss import borehole as pybh
from pysiss.utilities import Collection
import unittest


class TestBoreholeCollection(unittest.TestCase):

    """ Tests for BoreholeCollection
    """

    def setUp(self):
        self.boreholes = [pybh.Borehole("test_{0}".format(idx))
                          for idx in range(10)]
        self.bh_names = [bh.name for bh in self.boreholes]

    def test_creation(self):
        coll = Collection()
        self.assertEqual(len(coll), 0)
        self.assertEqual(len(coll._index), 0)

    def test_creation_2(self):
        """ Test initialization with a list of boreholes
        """
        coll = Collection(self.boreholes)
        for idx, (name, bh) in enumerate(coll.items()):
            self.assertEqual(bh, self.boreholes[idx])
            self.assertEqual(name, self.boreholes[idx].name)

    def test_addition(self):
        coll = Collection()
        for bh in self.boreholes:
            coll.append(bh)

        for idx, bh in enumerate(coll):
            self.assertEqual(bh, self.boreholes[idx])

    def test_iteration(self):
        coll = Collection(self.boreholes)
        for idx, bh in enumerate(coll):
            self.assertEqual(bh, self.boreholes[idx])

    def test_items(self):
        coll = Collection(self.boreholes)
        for idx, (name, bh) in enumerate(coll.items()):
            self.assertEqual(bh, self.boreholes[idx])
            self.assertEqual(name, self.boreholes[idx].name)



""" file:   __init__.py (pysiss)
    author: Jess Robertson
            CSIRO Minerals Resources Flagship
    date:   Monday 28 July 2014

    description: Init file for pysiss
"""

import borehole
import coverage
import vocabulary
import utilities

__all__ = [borehole, coverage, vocabulary, utilities]



""" file:   __init__.py (pysiss.vocabulary)
    author: Jess Robertson
            Mineral Resources Flagship, CSIRO
    date:   something

    description: init file
"""

import gsml
import gml
import erml
import namespaces

__all__ = [gsml, gml, erml, namespaces]



""" file:   __init__.py (pysiss.borehole.importers)
    author: Jess Robertson
            CSIRO Earth Science and Resource Engineering
    email:  jesse.robertson@csiro.au
    date:   Wednesday May 1, 2013

    description: Initialisation of the pysiss.borehole.importers module.
"""

import nvcl

__all__ = [nvcl]



""" file:   unmarshallers.py (pysiss.vocabulary.gml)
    author: Jess Robertson
            CSIRO Mineral Resources Flagship
    date:   Monday 25 August, 2014

    description: Unmarshalling functions for GeoSciML/GML objects
"""

from ...coverage.vector import MappedFeature
from ...metadata import Metadata
from ..namespaces import NamespaceRegistry, expand_namespace, shorten_namespace
from ..gml.unmarshallers import UNMARSHALLERS as GML_UNMARSHALLERS

NAMESPACES = NamespaceRegistry()


def mapped_feature(elem):
    """ Unmarshal a gsml:MappedFeature element
    """
    # Shape and projection data
    shape_elem = elem.find('./gsml:shape', namespaces=NAMESPACES)
    shape_data = shape(shape_elem)
    shape_elem.clear()  # Remove shape element from metadata

    # Identifier
    ident = elem.get(expand_namespace('gml:id')) or None

    # Get specification metadata records
    spec_elem = elem.find('./gsml:specification', namespaces=NAMESPACES)
    spec = specification(spec_elem)

    return MappedFeature(ident=ident, shape=shape_data['shape'],
                         projection=shape_data['projection'],
                         specification=spec)


def specification(elem):
    """ Unmarshall a gsml:specification element
    """
    # If we only have an xlink, this is just a pointer to another record
    xlink = elem.get(expand_namespace('xlink:href'))
    if xlink:
        # Just return the metadata key
        # We need to strip out the # from the link
        return xlink.lstrip('#')

    # Otherwise we need to create a new metadata record, then return the
    # relevant key
    else:
        spec_elem = elem.iterchildren().next()
        ident = spec_elem.get(expand_namespace('gml:id'))
        mdata = Metadata(ident=ident,
                         type=shorten_namespace(spec_elem.tag),
                         tree=spec_elem)
        return mdata.ident


def shape(elem):
    """ Unmarshal a gsml:shape element

        Here we just pass through to underlying gml shape data
    """
    child = elem[0]
    unmarshal = GML_UNMARSHALLERS[shorten_namespace(child.tag)]
    return unmarshal(child)


def get_value(elem):
    """ Unmashall an element containing a gsml:value element somewhere in its
        descendents.

        Returns the text value for a given element, stripping out children of
        the given element
    """
    return elem.find('.//gsml:value/text()', namespaces=NAMESPACES)


def cgi_termrange(elem):
    """ Unmarshal a gsml:CGI_TermRange element

        Return the value range for a given element
    """
    return map(get_value,
               elem.xpath('.//gsml:CGI_TermValue', namespaces=NAMESPACES))


def sampling_frame(elem):
    """ Unmarshal a gsml:samplingFrame element
    """
    return elem.get(expand_namespace('xlink:href'))


UNMARSHALLERS = {
    'gsml:shape': shape,
    'gsml:value': get_value,
    'gsml:CGI_TermValue': get_value,
    'gsml:CGI_TermRange': cgi_termrange,
    'gsml:preferredAge': get_value,
    'gsml:observationMethod': get_value,
    'gsml:positionalAccuracy': get_value,
    'gsml:samplingFrame': sampling_frame,
    'gsml:MappedFeature': mapped_feature,
    'gsml:specification': specification
}

__all__ = (UNMARSHALLERS,)



import unittest
from pysiss.vocabulary.namespaces import split_namespace, \
    shorten_namespace, expand_namespace


class TestXMLNamespaces(unittest.TestCase):

    """ Tests for XML namespace munging
    """

    def test_split_namespace(self):
        """ Test namespace splitting returns the right values
        """
        expected = ('urn:cgi:xmlns:CGI:GeoSciML:2.0', 'MappedFeature')
        self.assertEqual(
            split_namespace('urn:cgi:xmlns:CGI:GeoSciML:2.0:MappedFeature'),
            expected)
        self.assertEqual(
            split_namespace('{urn:cgi:xmlns:CGI:GeoSciML:2.0}MappedFeature'),
            expected)
        self.assertEqual(
            split_namespace('gsml:MappedFeature'),
            ('gsml', 'MappedFeature'))

    def test_shorten_namespace(self):
        """ Check namespace shortening
        """
        self.assertEqual(
            shorten_namespace('{http://www.opengis.net/gml}boundedBy'),
            'gml:boundedBy')

    def test_expand_namespace(self):
        """ Check namespace expansion
        """
        self.assertEqual(
            expand_namespace('gsml:MappedFeature'),
            '{urn:cgi:xmlns:CGI:GeoSciML:2.0}MappedFeature')
        self.assertEqual(
            expand_namespace('gsml:MappedFeature', form='rdf'),
            'urn:cgi:xmlns:CGI:GeoSciML:2.0:MappedFeature')


if __name__ == '__main__':
    unittest.main()



""" file:   __init__.py (pysiss.borehole)
    author: Jess Robertson
            CSIRO Earth Science and Resource Engineering
    email:  jesse.robertson@csiro.au
    date:   Wednesday May 1, 2013

    description: Initialisation of the pysiss.borehole module.
"""

from .borehole import Borehole, Feature
from .datasets import DataSet, PointDataSet, IntervalDataSet
from .properties import Property, PropertyType
from pysiss.borehole.siss.borehole_generator import SISSBoreholeGenerator
from . import plotting, analysis

__all__ = [Borehole, Feature,
           DataSet, PointDataSet, IntervalDataSet,
           Property, PropertyType,
           SISSBoreholeGenerator,
           plotting, analysis]



""" file:   plotting.py (pysiss.borehole)
    author: Jess Robertson
            CSIRO Earth Science and Resource Engineering
    email:  jesse.robertson@csiro.au
    date:   Wednesday May 1, 2013

    description: Plotting for the pysiss.borehole module.
"""

import matplotlib.pyplot
import matplotlib.cm
import matplotlib.collections
import numpy


def make_figure_grid(nplots, ncols=3, size=6):
    """ Make a grid of images
    """
    nrows = nplots / ncols
    if nrows * ncols != nplots:
        nrows += 1
    fig = matplotlib.pyplot.figure(figsize=(ncols * size, nrows * size))
    axeses = [matplotlib.pyplot.subplot(nrows, ncols, i + 1)
              for i in range(nplots)]
    return fig, axeses


def plot_variable(axes, variable, cmap=None):
    """ Generate a single image of a wavelet transform
    """
    if cmap is None:
        cmap = matplotlib.cm.get_cmap('Paired')
    axes.imshow(variable.T, cmap=cmap, interpolation='nearest')
    axes.set_aspect(variable.shape[0] / float(variable.shape[1]))
    axes.set_xlabel(r"Fourier Scale $\lambda_{i}$")
    axes.set_xticklabels([''])
    axes.set_xticks([])
    axes.set_ylabel(r"DataSet $x_{i}$")
    axes.set_yticklabels([''])
    axes.set_yticks([])


def plot_difference(axes, dataset, observed_value, expected_value,
                    colors=('red', 'blue'), orientation='horizontal'):
    """ Plots the difference between two series.

        This plot also includes a set of lines in the shading to indicate the
        sample spacing. You can set the colors that you want to use to
        highlight differences between the two series.

        :param axee: The axes to plot in
        :type axes: `matplotlib.pyplot.axes`
        :param dataset: The dataset variable
        :type dataset: `numpy.ndarray`
        :param observed_value: The actual value given as a numpy array. Must
            be the same length as the dataset vector.
        :type observed_value: `numpy.ndarray`
        :param expected_value: The expected value. Can be a constant, in which
            case the value will be constant, or an array of the same size as
            the dataset.
        :type expected_value: `numpy.ndarray` or number
        :param colors: A tuple of colors. The fill and lines will be shaded
            `color[0]` when `observed_value` < `expected_value` and `color[1]`
            when `observed_value` >= `expected_value`. Any color accepted by
            matplotlib is allowed.
        :type colors: Tuple
        :param orientation: The orientation of the plot, one of 'horizontal',
            or 'vertical'.
        :type orientation: str
    """
    # Expand the expected value if required
    if type(expected_value) is not numpy.ndarray:
        expected_value = expected_value * numpy.ones_like(dataset)

    # Helper functions for plotting the lines
    pos_diff = lambda a, b: a + numpy.maximum(b - a, 0)
    neg_diff = lambda a, b: a + numpy.minimum(b - a, 0)

    # Generate plot
    if orientation is 'horizontal':
        # Plot the two signals
        axes.plot(dataset, observed_value, color='black', linewidth=1)
        axes.plot(dataset, expected_value, 'k--', linewidth=1)

        # Generate fills
        axes.vlines(dataset, expected_value,
                    pos_diff(expected_value, observed_value),
                    color=colors[1])
        axes.vlines(dataset, neg_diff(expected_value, observed_value),
                    expected_value, color=colors[0])
        axes.fill_between(dataset, expected_value, observed_value,
                          where=(observed_value >= expected_value),
                          alpha=0.1,
                          facecolor=colors[1])
        axes.fill_between(dataset, expected_value, observed_value,
                          where=(observed_value < expected_value),
                          alpha=0.1,
                          facecolor=colors[0])
    elif orientation is 'vertical':
        # Plot the two signals
        axes.plot(observed_value, dataset, color='black', linewidth=2)
        axes.plot(expected_value, dataset, 'k--', linewidth=2)

        # Generate fills
        axes.hlines(dataset, expected_value,
                    pos_diff(expected_value, observed_value),
                    color=colors[1])
        axes.hlines(dataset, neg_diff(expected_value, observed_value),
                    expected_value, color=colors[0])
        axes.fill_betweenx(dataset, expected_value, observed_value,
                           where=(observed_value >= expected_value),
                           alpha=0.1,
                           facecolor=colors[1])
        axes.fill_betweenx(dataset, expected_value, observed_value,
                           where=(observed_value < expected_value),
                           alpha=0.1,
                           facecolor=colors[0])
    else:
        raise ValueError('Argument `orientation` must be "horizontal" or '
                         'vertical"')


def plot_signal(axes, dataset, signal, orientation='horizontal'):
    """ Plots a one-dimensional signal against some dataset.

        This assumes that you're plotting a detrended signal, so it fills in
        red for negative anomalies and blue for positive anomalies, so that
        you can compare the deviation from the trend.

        :param axes: The axes instance in which to plot the signal
        :type axes: `matplotlib.axes`
        :param dataset: An array of dataset locations
        :type dataset: `numpy.ndarray`
        :param signal: An array of signal values. Must be the same length as
            `dataset` or an error will be raised.
        :type signal: `numpy.ndarray`
        :param orientation: One of `'horizontal'` or `'vertical'`
        :type orientation: `str`
    """
    # Generate plot
    plot_difference(axes, dataset, signal, signal.mean(),
                    colors=('red', 'blue'),
                    orientation=orientation)
    if orientation is 'horizontal':
        axes.set_xlabel('DataSet $x$')
        axes.set_ylabel(r'Signal $f(x)$',
                        rotation=0,
                        horizontalalignment='right',
                        verticalalignment='center')
        axes.set_xlim(dataset[0], dataset[-1])
        axes.set_ylim(numpy.min(signal), numpy.max(signal))
    elif orientation is 'vertical':
        axes.set_ylabel('DataSet $x$')
        axes.set_xlabel('Signal $f(x)$')
        axes.set_xlim(numpy.min(signal), numpy.max(signal))
        axes.set_ylim(dataset[-1], dataset[0])
    else:
        raise ValueError('Argument `orientation` must be "horizontal" or '
                         'vertical"')


def plot_connection_graph(embedding, correlations, names, cluster_labels):
    """ Plots a connection graph in 2D given an embedding and a correlation
        matrix.
    """
    # Plot the nodes using the coordinates of our embedding
    axes = matplotlib.pyplot.gca()
    axes.scatter(embedding[0], embedding[1],
                 c=cluster_labels,
                 cmap=matplotlib.cm.get_cmap('Spectral'))

    # Plot the edges - a sequence of (*line0*, *line1*, *line2*), where
    #            linen = (x0, y0), (x1, y1), ... (xm, ym)
    non_zero = numpy.logical_not(correlations.mask)
    start_idx, end_idx = numpy.where(non_zero)
    segments = [[embedding[:, start], embedding[:, stop]]
                for start, stop in zip(start_idx, end_idx)]
    values = correlations[non_zero]
    lines = matplotlib.collections.LineCollection(
        segments,
        zorder=0,
        cmap=matplotlib.cm.get_cmap('RdBu'),
        norm=matplotlib.pylab.Normalize(.7 * values.min(), .7 * values.max()))
    lines.set_array(values)
    lines.set_linewidths(15 * numpy.abs(values))
    axes.add_collection(lines)

    # Add a label to each node
    label_info = zip(names, cluster_labels, embedding.T)
    for index, (name, label, (xloc, yloc)) in enumerate(label_info):
        (xloc, yloc), alignment = float_label(index, (xloc, yloc), embedding)
        point_color = matplotlib.cm.get_cmap('Spectral')(
            label / float(max(cluster_labels)))
        matplotlib.pylab.text(
            xloc, yloc, name, size=10,
            horizontalalignment=alignment[0],
            verticalalignment=alignment[1],
            bbox=dict(facecolor=point_color,
                      edgecolor=point_color,
                      alpha=.3))

    # Adjust axes limits
    axes.set_xlim(embedding[0].min() - .15 * embedding[0].ptp(),
                  embedding[0].max() + .10 * embedding[0].ptp(),)
    axes.set_ylim(embedding[1].min() - .03 * embedding[1].ptp(),
                  embedding[1].max() + .03 * embedding[1].ptp())


def float_label(index, position, embedding):
    """ Floating labels for plot so that they avoid one another.

        The challenge here is that we want to position the labels to avoid
        overlap with other labels. We use the neighbour data in the embedding
        to juggle the label positions to avoid collisions.
    """
    xloc, yloc = position
    dxloc = xloc - embedding[0]
    dxloc[index] = 1
    dyloc = yloc - embedding[1]
    dyloc[index] = 1
    this_dxloc = dxloc[numpy.argmin(numpy.abs(dyloc))]
    this_dyloc = dyloc[numpy.argmin(numpy.abs(dxloc))]
    if this_dxloc > 0:
        horizontalalignment = 'left'
        xloc += .002
        yloc += .001
    else:
        horizontalalignment = 'right'
        xloc -= .002
        yloc -= .001
    if this_dyloc > 0:
        verticalalignment = 'bottom'
        yloc += .002
        xloc += .001
    else:
        verticalalignment = 'top'
        yloc -= .002
        xloc -= .001

    return (xloc, yloc), (horizontalalignment, verticalalignment)


## Borehole plotting
def plot_point_dataset_data(point_dataset, keys_to_plot=None):
    """ Plot the data stored in the current node object

        :returns: handles to the figure and axes
    """
    if keys_to_plot is None:
        keys_to_plot = [
            k for k in point_dataset.properties.keys()
            if point_dataset.properties[k].property_type.isnumeric]

    # Plot data
    fig = matplotlib.pyplot.figure(figsize=(1 * len(keys_to_plot), 20))
    dataset_bounds = (point_dataset.depths.max(),
                      point_dataset.depths.min())
    for i, key in enumerate(keys_to_plot):
        axes = matplotlib.pyplot.subplot(1, len(keys_to_plot), i + 1)
        try:
            plot_signal(axes,
                        signal=point_dataset.properties[key].values,
                        dataset=point_dataset.depths,
                        orientation='vertical')
        except TypeError:
            print key
        axes.set_xlabel("")
        if i == 0:
            axes.set_ylabel('Depth (m)')
        else:
            axes.set_ylabel("")
            axes.set_yticklabels("")
        axes.set_ylim(dataset_bounds)
        axes.xaxis.set_major_locator(matplotlib.pyplot.MaxNLocator(3))
        axes.set_title(point_dataset.properties[key].property_type.long_name,
                       rotation=90,
                       verticalalignment='bottom',
                       horizontalalignment='center')
    fig.tight_layout()
    return fig, axes


def gen_axes_grid(nplots, ncols):
    """ Make an axes grid with the given number of columns and plots
    """
    nrows = nplots // ncols
    if nrows * ncols != nplots:
        nrows += 1
    return matplotlib.gridspec.GridSpec(nrows, ncols)



""" file:   utilities.py (pysiss.borehole)
    author: Jess Robertson
            CSIRO Earth Science and Resource Engineering
    email:  jesse.robertson@csiro.au
    date:   Wednesday May 1, 2013

    description: Utility functions for the cwavelet module.
"""

from maths import *
from collection import Collection
from id_object import id_object
# from projection import project
from singleton import Singleton



""" file: slow_decorator.py
    author: Jess Robertson
            CSIRO Minerals Resources Flagship
    date:   Thursday 29 May, 2014

    description: Adds a facility to skip slow tests
"""

import unittest

# Set the flag below to False to run the slow tests
SKIP_SLOW = True


# Define a custom unittest decorator to tag slow tests which should be skipped
def slow(obj):
    """ Decorator to skip slow tests
    """
    if SKIP_SLOW:
        return unittest.skip('Skipping slow tests')(obj)
    else:
        return obj



""" file:   __init__.py (pysiss.metadata)
    author: Jess Robertson
            CSIRO Minerals Resources Flagship
    date:   Wednesday 27 August, 2014

    description: Functions to deal with metadata
"""

from .registry import MetadataRegistry
from .metadata import Metadata

__all__ = [MetadataRegistry, Metadata]



""" file: __init__.py (pysiss.borehole.datasets)
    author: Jess Robertson
            CSIRO Earth Science and Resource Engineering
    date: Sunday November 10, 2013

    description: Imports for pyborholes.datasets
"""

from .dataset import DataSet
from .point_dataset import PointDataSet
from .interval_dataset import IntervalDataSet



""" file:   details.py (pysiss.borehole)
    author: Jess Robertson & Ben Caradoc-Davies
            CSIRO Earth Science and Resource Engineering
    date:   September 10, 2013

    description: Details class implementation
"""

from collections import namedtuple


def detail_type(detail_name, detail_attr=None):
    """ Return a detail type for use in a Details class

        This is essentially a wrapper around `collections.namedtuple`,
        for more details on the formatting of the detail_attr, see the
        Python documentation.

        :param detail_name: The identifier of the new detail type
        :type detail_name: string
        :param detail_attr: The attributes of the new detail type
        :type detail_attr: string
    """
    if detail_attr is None:
        detail_attr = 'name values property_type'
    return namedtuple(detail_name, detail_attr)


class Details(dict):

    """ Class to store details about another class
    """

    def __init__(self):
        super(Details, self).__init__()

    def add_detail(self, name, values, property_type=None):
        """ Add a detail to the class

            :param name: An identifier for the detail
            :type name: string
            :param values: The data to add
            :type values: any Python object
            :param property_type: The property type of the detail, optional,
                defaults to None
            :type property_type: `pysiss.borehole.PropertyType`
        """
        self[name] = self.detail_type(name=name,
                                      values=values,
                                      property_type=property_type)

    def __setattr__(self):
        """ Disable setattr method
        """
        raise NotImplementedError('Use add_detail to add details')



""" file:   __init__.py (pysiss.borehole.siss)
    author: David Benn
            CSIRO IM&T Scientific Computing Data Processing Services
    email:  David.Benn@csiro.au
    date:   Tuesday May 20, 2014

    description: Initialisation of the pysiss.borehole.siss module.
"""

import borehole_generator


""" file: dataset.py (pysiss.borehole.datasets)
    author: Jess Robertson
            CSIRO Earth Science and Resource Engineering
    date: Sunday November 10, 2013

    description: Classes to represent data datasets etc in boreholes

    This file defines the abstract base class DataSet which acts as a parent
    or all the dataset data types; it should not be instantiated by users.
"""

from ..properties import Property
from ..details import Details, detail_type
from ...utilities import id_object


class DataSet(id_object):

    """ Spatial extent over which properties are defined.

        This is an abstract base class

        Some important properties are:
            properties - dict mapping property name to Property
            size - the size of all the values sequences
            name - an identifier
            subdatasets - a list of subdataset locations
            gaps - a list of gap locations
            details - the metadata associated with this dataset

        :param name: an identifier for the dataset
        :type name: string
        :param size: The number of items in the dataset
        :type size: int
        :param details: Metadata for a given dataset.
        :type details: pysiss.borehole.datasets.DatasetDetails
    """

    def __init__(self, name, size, details=None):
        assert size > 0, "dataset must have at least one element"
        self.properties = dict()
        self.size = size  # size of all values sequences
        self.name = name
        self.subdatasets = None
        self.gaps = None
        self.details = details

    def add_property(self, property_type, values):
        """ Add and return a new property
        """
        assert self.size == len(values), ("values must have the same number "
                                          "of elements as the dataset")
        self.properties[property_type.name] = Property(property_type, values)
        return self.properties[property_type.name]

    def get_property_names(self):
        """ Return the properties defined over this dataset
        """
        return self.properties.keys()

    def to_dataframe(self):
        """ Tranform the data in the dataset into a Pandas dataframe.
        """
        raise NotImplemented


class DatasetDetails(Details):

    """ Class to store metadata about a dataset
    """

    detail_type = detail_type('BoreholeDetail', 'name values property_type')



#!/usr/bin/env python
""" file:   test_borehole.py
    author: Ben Caradoc-Davies
            CSIRO Earth Science and Resource ENgineering
    date:   November 20, 2013

    description: Unit tests to demonstrate the borehole model.
"""

from pysiss import borehole as pybh
import numpy
import unittest

DENSITY = pybh.PropertyType(name="d",
                            long_name="density",
                            units="g/cm3")

# bogus unit from original spreadsheet
IMPEDANCE = pybh.PropertyType(name="imp",
                              long_name="impedance",
                              units="kg/m2.s.10-3")

ROCK_TYPE = pybh.PropertyType(name="rock", long_name="rock type")


class BoreholeTest(unittest.TestCase):

    def setUp(self):
        self.borehole = pybh.Borehole("test")

    def test_name(self):
        """ Test that name is being captured for borehole
        """
        self.assertEquals(self.borehole.name, "test")

    def test_features(self):
        """ Test store and retrieve a single point feature with a one
            single-valued category property
        """
        feature = self.borehole.add_feature("fault-1", 27.3)
        feature.add_property(pybh.PropertyType("age"), "last friday")
        self.assertEquals(
            "last friday",
            self.borehole.features["fault-1"].properties["age"].values)

    def test_interval_dataset(self):
        """ Test store and retrieve four intervals with a single multivalued
            category property
        """
        geology_intervals = [
            [7.0, 8.0, "SA"],
            [8.0, 11.0, "SL", "CA"],
            [11.0, 14.0, "SC", "CA"],
            [14.0, 15.0, "SC", "FE"]
        ]
        # the starts and ends of the intervals
        from_depths = numpy.asarray([x[0] for x in geology_intervals])
        to_depths = numpy.asarray([x[1] for x in geology_intervals])
        # rock type is a multivalued category property
        rock_type = [x[2:] for x in geology_intervals]
        self.assertTrue(all(numpy.diff(from_depths) > 0))
        self.assertTrue(all(numpy.diff(to_depths) > 0))
        dataset = self.borehole.add_interval_dataset("geology",
                                                     from_depths,
                                                     to_depths)
        dataset.add_property(ROCK_TYPE, rock_type)
        geol = self.borehole.interval_datasets["geology"]
        self.assertTrue(all(from_depths == geol.from_depths))
        self.assertTrue(all(to_depths == geol.to_depths))
        self.assertEquals(["SC", "FE"], geol.properties["rock"].values[-1])

    def test_point_datasets(self):
        """Test store and retrieve two properties sampled at four depths"""
        # the depths of the dataset
        depths = [4.0, 4.02, 4.0603, 4.0803]
        # two numerical properties
        densities = [2.8073, 2.837, 2.8569, 2.8158]
        impedances = [9010.898, 9250.686, 11854.32, 11621.28]
        dataset = self.borehole.add_point_dataset("samples", depths)
        dataset.add_property(DENSITY, densities)
        dataset.add_property(IMPEDANCE, impedances)
        self.assertTrue(
            all(depths == self.borehole.point_datasets["samples"].depths))
        self.assertEquals(
            densities,
            self.borehole.point_datasets["samples"].properties["d"].values)
        self.assertEquals(
            impedances,
            self.borehole.point_datasets["samples"].properties["imp"].values)
        self.assertEquals(
            "samples",
            self.borehole.point_datasets['samples'].name)

    def test_interval_dataset_depths_empty(self):
        """ Test that empty interval depths raises an AssertionError
        """
        self.assertRaises(
            AssertionError,
            lambda: pybh.IntervalDataSet("test", [], []))

    def test_interval_dataset_depths_different_size(self):
        """ Test that interval depths of different size raises an
            AssertionError
        """
        self.assertRaises(
            AssertionError,
            lambda: pybh.IntervalDataSet("test", [1, 3, 5], [2, 4]))

    def test_interval_dataset_depths_decreasing(self):
        """ Test that decreasing interval depths raises an AssertionError
        """
        self.assertRaises(
            AssertionError,
            lambda: pybh.IntervalDataSet("test", [3, 1], [4, 2]))

    def test_interval_dataset_zero_length(self):
        """ Test that zero interval length raises an AssertionError
        """
        self.assertRaises(
            AssertionError,
            lambda: pybh.IntervalDataSet("test", [1, 3], [2, 3]))

    def test_interval_dataset_negative_length(self):
        """ Test that negative interval length raises an AssertionError
        """
        self.assertRaises(
            AssertionError,
            lambda: pybh.IntervalDataSet("test", [1, 4], [2, 3]))

    def test_interval_dataset_overlap(self):
        """ Test that interval overlap raises an AssertionError
        """
        self.assertRaises(
            AssertionError,
            lambda: pybh.IntervalDataSet("test", [1, 2], [3, 4]))

    def test_interval_dataset_wrong_property_size(self):
        """ Test that interval property with wrong number of values raises an
            AssertionError
        """
        dset = pybh.IntervalDataSet("test", [1, 3], [2, 4])
        self.assertRaises(
            AssertionError,
            lambda: dset.add_property(DENSITY, [1.3]))

    def test_point_dataset_depths_empty(self):
        """ Test that empty point depths raises an AssertionError
        """
        self.assertRaises(
            AssertionError,
            lambda: pybh.PointDataSet("test", []))

    def test_sampling_dataset_depths_decreasing(self):
        """ Test that decreasing point depths raises an AssertionError
        """
        self.assertRaises(
            AssertionError,
            lambda: pybh.PointDataSet("test", [2, 1]))

    def test_sampling_dataset_wrong_property_size(self):
        """ Test that point property with wrong number of values raises an
            AssertionError
        """
        dset = pybh.PointDataSet("test", [1, 2])
        self.assertRaises(
            AssertionError,
            lambda: dset.add_property(DENSITY, [1.3]))


if __name__ == "__main__":
    unittest.main()



""" file: test_siss.py

    author: David Benn
            CSIRO IM&T Scientific Computing Data Processing Services
    date: 18 February 2014

    description: Test of siss module for creating Borehole objects populated
                 with GeoSciML metadata as borehole details.
"""

from datetime import datetime
import os
import unittest
import requests

import pysiss.borehole as pybh


class SissTest(unittest.TestCase):

    def setUp(self):
        self.siss = pybh.SISSBoreholeGenerator()
        self.test_dir = os.path.dirname(os.path.realpath(__file__))

    def tearDown(self):
        if os.path.exists('test_response_file.xml'):
            os.remove('test_response_file.xml')

    def test_geosciml_2_borehole(self):
        """ A test of GeoSciML 2.0 handling in SISSBoreholeGenerator.
            Here we read a test XML file.
        """
        xml_file = '{0}/geosciml/geo2test.xml'.format(self.test_dir)
        bh = self.siss.geosciml_to_borehole('geosciml2_test', xml_file)

        self.assertEquals(-29.804238 * self.siss.unit_reg.degree,
                          bh.origin_position.latitude)
        self.assertEquals(139.007411 * self.siss.unit_reg.degree,
                          bh.origin_position.longitude)
        self.assertEquals(0.0 * self.siss.unit_reg.meter,
                          bh.origin_position.elevation)
        self.assertEquals('elevation: Gravity-related height',
                          bh.origin_position.property_type.description)

        self.assertEquals('DMITRE', bh.details.get('driller').values)
        self.assertEquals('Diamond', bh.details.get('drilling method').values)
        self.assertEquals(datetime(year=2009, month=7, day=27),
                          bh.details.get('date of drilling').values)
        self.assertEquals('natural ground surface',
                          bh.details.get('start point').values)
        self.assertEquals('vertical',
                          bh.details.get('inclination type').values)
        self.assertEquals([90.0, 0.0, 90.0, 0.0],
                          bh.details.get('shape').values)
        self.assertEquals({'lower corner': 279.0 * self.siss.unit_reg.meter,
                           'upper corner': 351.29 * self.siss.unit_reg.meter},
                          bh.details.get('cored interval').values)
        self.assertEquals(1 * self.siss.unit_reg.meter,
                          bh.details.get('cored interval').property_type.units)

    def test_geosciml_3_borehole(self):
        """ A test of GeoSciML 3.0 handling in SISSBoreholeGenerator.
            Here we read a test XML file.
        """
        xml_file = "{0}/geosciml/geo3test.xml".format(self.test_dir)
        bh = self.siss.geosciml_to_borehole("geosciml3_test", xml_file)

        self.assertEquals(-25.0 * self.siss.unit_reg.degree,
                          bh.origin_position.latitude)
        self.assertEquals(119.0 * self.siss.unit_reg.degree,
                          bh.origin_position.longitude)
        self.assertEquals(210.0 * self.siss.unit_reg.meter,
                          bh.origin_position.elevation)
        self.assertEquals('description: Rotary Table Position',
                          bh.origin_position.property_type.description)

        self.assertEquals('Gelogical Survey of Finland',
                          bh.details.get('driller').values)
        self.assertEquals('diamond core',
                          bh.details.get('drilling method').values)
        self.assertEquals(datetime(year=1984, month=5, day=15),
                          bh.details.get('date of drilling').values)
        self.assertEquals('natural ground surface',
                          bh.details.get('start point').values)
        self.assertEquals('inclined down',
                          bh.details.get('inclination type').values)
        self.assertEquals([-25.0, 119.0, 210, -25.0, 119.01, 177],
                          bh.details.get('shape').values)
        self.assertEquals({'lower corner': 0,
                           'upper corner': 125.0},
                          bh.details.get('cored interval').values)

    def test_geosciml_nvcl_scanned_borehole(self):
        """ A test using an XML document corresponding to a scanned
            borehole GeoSciML URL.
            Note: this test has a dependency upon a NVCL web resource existing!
        """
        bh_name = 'WTB5'
        bh_url = ('http://nvclwebservices.vm.csiro.au/resource/'
                  'feature/CSIRO/borehole/WTB5')
        response = requests.get(bh_url)
        self.assertTrue(response.ok)
        with open('test_response_file.xml', 'wb') as fhandle:
            fhandle.write(response.content)
        bh = self.siss.geosciml_to_borehole(bh_name, 'test_response_file.xml')

        self.assertEquals(-28.4139 * self.siss.unit_reg.degree,
                          bh.origin_position.latitude)
        self.assertEquals(121.142 * self.siss.unit_reg.degree,
                          bh.origin_position.longitude)
        self.assertEquals(45.0 * self.siss.unit_reg.meter,
                          bh.origin_position.elevation)
        self.assertEquals('elevation: Gravity-related height',
                          bh.origin_position.property_type.description)

        self.assertEquals('GSWA', bh.details.get('driller').values)
        self.assertEquals('diamond core',
                          bh.details.get('drilling method').values)
        self.assertEquals(datetime(year=2004, month=9, day=17),
                          bh.details.get('date of drilling').values)
        self.assertEquals('natural ground surface',
                          bh.details.get('start point').values)
        self.assertEquals('vertical',
                          bh.details.get('inclination type').values)
        self.assertEquals([-28.4139, 121.142, -28.4139, 121.142],
                          bh.details.get('shape').values)
        self.assertEquals({'lower corner': 106.0 * self.siss.unit_reg.meter,
                           'upper corner': 249.0 * self.siss.unit_reg.meter},
                          bh.details.get('cored interval').values)
        self.assertEquals(1 * self.siss.unit_reg.meter,
                          bh.details.get('cored interval').property_type.units)



""" file:   registry.py (pysiss.metadata)
    author: Jess Robertson
            CSIRO Minerals Resources Flagship
    date:   Wednesday 27 August, 2014

    description: Functions to deal with metadata objects

    Metadata descriptions can be shared by many different objects, so it makes
    sense to seperate these out into a seperate registry.
"""

from ..utilities import Singleton


class MetadataRegistry(dict):

    """ A registry to store metadata instances

        Since GeoSciML allows metadata reuse, we need to have a central
        repository of metadata which stores the actual etrees, and objects can
        refer to keys within this repository.
    """

    __metaclass__ = Singleton

    def register(self, metadata_item):
        """ Register a metadata item in the registry
        """
        self[metadata_item.ident] = metadata_item

    def deregister(self, metadata_key):
        """ Deregister the given metadata item given by the key
        """
        del self[metadata_key]



""" file: nvcl.py (pysiss.borehole.importers)
    author: Josh Vote and Jess Robertson
            CSIRO Earth Science and Resource Engineering
    date:   23 December 2013

    description: Importer for NVCL data services
"""

from ..borehole import PropertyType, SISSBoreholeGenerator
from ..borehole.datasets import PointDataSet  # , IntervalDataSet
from ..utilities import Singleton

from owslib.wfs import WebFeatureService
import numpy
import pandas
import requests
from lxml import etree
from StringIO import StringIO


NVCL_DEFAULT_ENDPOINTS = {
    'CSIRO': {
        'wfsurl': 'http://nvclwebservices.vm.csiro.au/geoserverBH/wfs',
        'dataurl': 'http://nvclwebservices.vm.csiro.au/NVCLDataServices/',
        'downloadurl':
        'http://nvclwebservices.vm.csiro.au/NVCLDownloadServices/'
    },
    'GSWA': {
        'wfsurl': 'http://geossdi.dmp.wa.gov.au/services/wfs',
        'dataurl': 'http://geossdi.dmp.wa.gov.au/NVCLDataServices/',
        'downloadurl': 'http://geossdi.dmp.wa.gov.au/NVCLDownloadServices/'
    },
    'MRT': {
        'wfsurl': 'http://www.mrt.tas.gov.au/web-services/wfs',
        'dataurl': 'http://www.mrt.tas.gov.au/NVCLDataServices/',
        'downloadurl': None
    },
    'NTGS': {
        'wfsurl': 'http://geology.data.nt.gov.au/nvcl/wfs',
        'dataurl': 'http://geology.data.nt.gov.au/NVCLDataServices/',
        'downloadurl': 'http://geology.data.nt.gov.au/NVCLDownloadServices/'
    },
    'PIRSA': {
        'wfsurl': 'https://egate.pir.sa.gov.au/nvcl/geoserver/wfs',
        'dataurl': 'https://egate.pir.sa.gov.au/nvcl/NVCLDataServices/',
        'downloadurl': 'https://egate.pir.sa.gov.au/nvcl/NVCLDownloadServices/'
    },
    'DPINSW': {
        'wfsurl': 'http://auscope.dpi.nsw.gov.au/geoserver/wfs',
        'dataurl': 'http://auscope.dpi.nsw.gov.au/NVCLDataServices/',
        'downloadurl': 'http://auscope.dpi.nsw.gov.au/NVCLDownloadServices/'
    }
}


class NVCLEndpointRegistry(dict):

    """ Registry to manage information about NVCL endpoints

        Each endpoint stores a WFS URL, a data URL for queries about available
        data, and a download URL for getting analyte data. URLS are retrieved
        using standard dictionary syntax:

            registry = NVCLEndpointRegistry()
            url = registry['CSIRO']['wfsurl']  # Get the WFS URL for the CSIRO
                                               # NVCL endpoint

        New endpoints can be registered using `NVCLEndpointRegistry.register`.
    """

    __metaclass__ = Singleton

    def __init__(self):
        for endpoint, urls in NVCL_DEFAULT_ENDPOINTS.items():
            self.register(endpoint, **urls)

    def register(self, endpoint, wfsurl=None, dataurl=None, downloadurl=None,
                 update=False):
        """ Register an NVCL endpoint in the registry.

            This method avoids clobbering existing endpoints. If you want to
            update an existing endpoint, set the update flag to True.

            :param endpoint: A unique identifier for the endpoint
            :type endpoint: string
            :param wfsurl: The URL pointing to the web feature service for the
                NVCL endpoint.
            :type wfsurl: string
            :param dataurl: The URL pointing to the data query service for the
                NVCL endpoint.
            :type dataurl: string
            :param downloadurl: The URL pointing to the data download service
                for the NVCL endpoint.
            :type downloadurl: string

        """
        # Check that we don't already have this endpoint registered
        if endpoint in self.keys() and not(update):
            raise KeyError("Endpoint {0} is already in the registry and you "
                           "don't want to update it".format(endpoint))

        # Add the endpoint to the registry
        self[endpoint] = {
            'wfsurl': wfsurl,
            'dataurl': dataurl,
            'downloadurl': downloadurl
        }


class NVCLImporter(object):

    """ Import boreholes by consuming NVCL services

        :param endpoint: An endpoint identifier. To get a list of currently
            registered endpoints, call `NVCLEndpointRegistry().keys()`. A
            KeyError is raised if an unknown endpoint is used.
        :type endpoint: string
    """

    def __init__(self, endpoint='CSIRO'):
        super(NVCLImporter, self).__init__()
        self.endpoint = endpoint

        # Get URL data associated with endpoint
        registry = NVCLEndpointRegistry()
        try:
            self.urls = registry[endpoint]
        except KeyError:
            raise KeyError('Unknown NVCL Endpoint {0}.'.format(endpoint) +
                           'Registered endpoints: {0}'.format(registry.keys()))

        # Set up SISSBoreholeGenerator instance
        # TODO: use this in get_borehole() instead of creating another
        # instance?
        self.generator = SISSBoreholeGenerator()

    def __repr__(self):
        """ String representation
        """
        str = 'NVCLImporter(endpoint="{0}")'.format(self.endpoint)
        return str

    def get_borehole_idents_and_urls(self, maxids=None):
        """ Generates a dictionary containing identifiers and urls for
            boreholes with NVCL scanned data at this endpoint

            :param maxids: The maximum number of boreholes to request or
                None for no limit
            :type maxids: integer
            :returns: an dictionary of urls keyed by borehole identifiers
        """
        wfs = WebFeatureService(self.urls['wfsurl'], version="1.1.0")
        wfsresponse = wfs.getfeature(
            typename="nvcl:ScannedBoreholeCollection",
            maxfeatures=maxids)
        xmltree = etree.parse(wfsresponse)

        idents = {}
        bhstring = ".//{http://www.auscope.org/nvcl}scannedBorehole"
        for match in xmltree.findall(bhstring):
            idents[match.get('{http://www.w3.org/1999/xlink}title')] = \
                match.get('{http://www.w3.org/1999/xlink}href')
        return idents

    def get_borehole_idents(self, maxids=None):
        """ Returns the identifiers of boreholes with NVCL scanned data

            :param maxids: The maximum number of boreholes to request or
                None for no limit
            :type maxids: integer
            :returns: a list borehole identifiers
        """
        return self.get_borehole_idents_and_urls(maxids).keys()

    def get_dataset_idents(self, hole_ident):
        """ Generates a dictionary of tuples representing all the NVCL datasets
            associated with this particular borehole

            :param hole_ident: The GUID for a borehole available at dataurl
            :type hole_ident: string
            :returns: a dictionary keyed by dataset name, where each dictionary
                value is the GUID of the dataset.
        """
        xmltree = None
        holeurl = (self.urls['dataurl'] + 'getDatasetCollection.html?'
                   'holeidentifier={0}').format(hole_ident)
        response = requests.get(holeurl)
        if response:
            xmltree = etree.fromstring(response.content)

            datasets = {}
            for dset in xmltree.findall(".//Dataset"):
                datasets[dset.find('DatasetName').text] = \
                    dset.find('DatasetID').text
            return datasets

    def get_analyte_idents(self, hole_ident, dataset_ident):
        """ Generates a dictionary mapping all NVCL analytes for a given
            borehole dataset to their GUIDs.

            Returns None if the XML fails to parse.

            :param hole_ident: The GUID for a borehole available at dataurl
            :type hole_ident: string
            :param dataset_ident: The GUID for a dataset available at dataurl
            :type dataset_ident: string
            :returns: a dictionary keyed by analyte name, where each value is
                the GUID for a given analyte.
        """
        analyte_idents = None
        dseturl = 'getLogCollection.html?mosaicsvc=no&datasetid={0}'
        response = requests.get(self.urls['dataurl']
                                + dseturl.format(dataset_ident))

        # Parse XML tree to return analytes
        if response:
            xmltree = etree.fromstring(response.content)
            analyte_idents = {}
            for analyte in xmltree.findall(".//Log"):
                log_ident = analyte.find("LogID").text
                name = analyte.find("logName").text
                # sample_count = analyte.find('SampleCount').text
                analyte_idents[name] = log_ident
            return analyte_idents
        else:
            raise Exception(
                'Request for data returned {0}'.format(response.code))

    def get_analytes(self, hole_ident, dataset_name, dataset_ident,
                     analyte_idents=None,
                     from_depth=None, to_depth=None,):
        """ Get the analytes from the given borehole and dataset

            :param hole_ident: The identifier for a borehole
            :type hole_ident: string
            :param dataset_name: An identifier for the generated dataset
            :type dataset_name: string
            :param dataset_ident: The identifier for the dataset
            :type dataset_ident: string
            :param analyte_idents: The identifers of the analytes to download.
                Optional, if None then all analytes in the given dataset are
                downloaded.
            :type analyte_idents: list of str
            :param from_depth/to_depth: The depth range included in the
                dataset. Optional, defaults to the entire depths defined in
                the NVCL.
            :type from_depth/to_depth: float
        """
        # Get analyte data
        analyte_ident_dict = self.get_analyte_idents(hole_ident, dataset_ident)
        if len(analyte_ident_dict) == 0:
            # This dataset has no analytes
            print 'Warning, dataset {0} has no analytes'.format(dataset_ident)
            return None

        # Generate request URL
        if analyte_idents is None:
            analyte_idents = analyte_ident_dict.values()
        url = self.urls['dataurl'] + 'downloadscalars.html?'
        for ident in analyte_idents:
            url += '&logid={0}'.format(ident)

        # We'll use pandas to slurp the csv direct from the web service
        analytedata = pandas.read_csv(url)
        startcol = 'StartDepth'
        endcol = 'EndDepth'
        analytecols = [k for k in analytedata.keys()
                       if k not in (startcol, endcol)]

        # NVCL data results in start depths == end depths.
        # Ranges aren't really appropriate. Better to use sampling
        # dataset
        analytedata = analytedata.drop_duplicates(startcol)
        startdepths = numpy.asarray(analytedata[startcol])
        dataset = PointDataSet(dataset_name, startdepths)

        # Make a property for each analyte in the borehole
        #
        # TODO: What to do with this? Despite now having borehole
        #       details, we still need to make a correspondence
        #       between analyte data and the borehole. Is what
        #       follows still valid?
        #
        for analyte in analytecols:
            property_type = PropertyType(
                name=analyte,
                long_name=analyte,
                units=None,
                description=None,
                isnumeric=False)
            dataset.add_property(
                property_type=property_type,
                values=numpy.asarray(analytedata[analyte]))

        return dataset

    def get_mosaic(self, hole_ident, from_depth=None, to_depth=None):
        """ Requests a mosaic from the NVCL data portal

            The mosaic is a low-resolution composite of the image data
            associated with a given borehole, so is suitable for large
            borehole ranges
        """
        raise NotImplemented

    def get_images(self, hole_ident, from_depth=None, to_depth=None):
        """ Requests high-resolution images from the NVCL data portal

            These images are high-resolution and represent slices of the
            core sitting in the core tray.
        """
        raise NotImplemented

    def get_borehole(self, hole_ident, name=None, get_analytes=True,
                     raise_error=True):
        """ Generates a pysiss.borehole.Borehole instance containing the data
            from the given borehole.

            Each dataset defined on the borehole is downloaded down into a
            seperate Dataset instance

            :param hole_ident: The hole identifier
            :type hole_ident: string
            :param name: Descriptive name for this borehole. Optional, if not
                specified this will default to the NVCL borehole id.
            :type name: string
            :param get_analytes: If True, the analytes will also be downloaded
            :type get_analytes: bool
            :param raise_error: Whether to raise an exception on an HTTP error
                (e.g. 404'd). If false, get_borehole returns None.
            :returns: a `pysiss.borehole.Borehole` object
        """
        try:
            # Generate pysiss.borehole.Borehole instance to hold the data
            if name is None:
                name = hole_ident
            siss_bhl_generator = SISSBoreholeGenerator()
            bh_url = self.get_borehole_idents_and_urls()[hole_ident]
            response = requests.get(bh_url)

            # Break out now if the request fails
            if not response:
                return None
            bhl = siss_bhl_generator.geosciml_to_borehole(
                name, StringIO(response.content))

            # For each dataset in the NVCL we want to add a dataset and store
            # the dataset information in the DatasetDetails
            if get_analytes:
                datasets = self.get_dataset_idents(hole_ident)
                for dataset_name, dataset_guid in datasets.items():
                    dataset = self.get_analytes(hole_ident=hole_ident,
                                                dataset_name=dataset_name,
                                                dataset_ident=dataset_guid)
                    if dataset is not None:
                        bhl.add_dataset(dataset)

            return bhl

        except Exception, err:
            if raise_error:
                raise err
            else:
                return None



""" file: property_type.py (pysiss.borehole.properties)
    author: Jess Robertson
            CSIRO Earth Science and Resource Engineering
    date: Sunday November 10, 2013

    description: Imports for pysiss.borehole.properties
"""

class PropertyType(object):

    """ The metadata for a property.

        :param name: Property name
        :type name: string
        :param long_name: Name for presentation to the user
        :type long: atring or `None`
        :param description: Descriptive phrase
        :type description: string
        :param units: Unit in Unified Code for Units of Measure (UCUM)
        :type units: string or None
        :param isnumeric: Whether the property is numeric or categorical
        :type isnumeric: bool
    """

    def __init__(self, name, long_name=None, description=None, units=None,
        isnumeric=True, detection_limit=None):
        self.name = name
        self._long_name = long_name
        self.description = description
        self.units = units
        self.isnumeric = isnumeric
        self.detection_limit = detection_limit

    def __repr__(self):
        info = 'PropertyType {0}: long name is "{1}", units are {2}'
        return info.format(self.name, self.long_name, self.units)

    @property
    def long_name(self):
        """ Return long name or name if no long name.
        """
        return self._long_name if self._long_name is not None else self.name

    def copy(self):
        """ Return a copy of the PropertyType instance
        """
        return PropertyType(self.name, self.long_name, self.description,
            self.units)




#!/usr/bin/env python
""" file: test_real_data.py
    author: Jess Robertson
            CSIRO Earth Science and Resource Engineering
    date:   Wednesday November 20, 2013

    description: Test running pysiss.borehole over a real dataset. This is
    bascially just a couple of scripts pulled from the example sets.
"""

import pysiss
import pandas
import numpy
import cPickle as pickle
import itertools
import os
import unittest

# Location of data files
PROP_TYPE_PKL = os.path.join('.', 'test_real_ptypes.pkl')
TEST_DATA_PKL = os.path.join('.', 'test_real_data.pkl')
DATA_FILE = os.path.join(
    '..', 'examples',
    'Vogue_Apr_2013_combined_pXR__ME_data_classified.csv')


def make_property_types():
    """ Makes property type instances
    """
    try:
        with open(PROP_TYPE_PKL, 'rb') as pkl_file:
            return pickle.load(pkl_file)
    except IOError:
        # We'll just keep going and generate the pickle by hand
        pass

    ## INFORMATION ABOUT DATAFILE
    # Labels and long-form version
    vogue_data = pandas.read_csv(DATA_FILE)
    long_names = dict(zip(vogue_data.keys(), itertools.repeat(None)))
    long_names.update({
        'As_ppm': 'Arsenic',
        'CaO_pct': 'Calcium',
        'Co_ppm': 'Cobalt',
        'Cr_ppm': 'Chromium',
        'Cu_ppm': 'Copper',
        'Fe2O3_pct': 'Ferric Iron (3+)',
        'K2O_pct': 'Potassium',
        'MnO_pct': 'Manganese',
        'Ni_ppm': 'Nickel',
        'Rb_ppm': 'Rubidium',
        'SO3_pct': 'Sulphur',
        'Sr_ppm': 'Strontium',
        'TiO2_pct': 'Titanium',
        'V_ppm': 'Vanadium',
        'Zn_ppm': 'Zinc',
        'Zr_ppm': 'Zirconium',
        'Sample': 'Sample number',
        'Hole_ID': 'Borehole ID',
        'From': 'Depth from',
        'To': 'Depth to',
        'TYPE': 'Measurement method',
        'Litho2': 'Lithology label'
    })

    # Set units
    def units(key):
        """ Returns the units for the given key
        """
        if 'ppm' in key:
            return 'ppm'
        elif 'pct' in key and 'O' in key:
            return 'oxide wt. %'
        elif key in ['From', 'To'] or 'mid' in key:
            return 'metres'
        else:
            return None

    def is_numeric(key):
        """ Returns whether data is numeric or not
        """
        numeric_labels = ['ppm', 'pct', 'Litho2']
        return any([l in key for l in numeric_labels])

    # Generate property_types
    property_types = dict()
    for key in vogue_data.keys():
        isnumeric = is_numeric(key)
        property_types[key] = pysiss.borehole.PropertyType(
            name=key, long_name=long_names[key], units=units(key),
            isnumeric=isnumeric)

    # Dump property types to pickle file and return
    with open(PROP_TYPE_PKL, 'wb') as pkl_file:
        pickle.dump(property_types, pkl_file)
    return property_types


class TestRealData(unittest.TestCase):

    """ Test running pysiss.borehole over a real dataset
    """

    def setUp(self):
        # Reload prewarmed data if available, else load data & cache
        property_types = make_property_types()
        try:
            pkl_file = open(self.test_data_pkl, 'rb')
            self.boreholes = pickle.load(pkl_file)
            return
        except IOError:
            pass

        # If we're here, we need to generate the data from the datafile
        vogue_data = pandas.read_csv(DATA_FILE)
        vogue_data = vogue_data.sort(('Hole_ID', 'From', 'To'))
        property_keys = [
            'As_ppm', 'CaO_pct', 'Co_ppm', 'Cr_ppm', 'Cu_ppm',
            'Fe2O3_pct', 'K2O_pct', 'MnO_pct', 'Ni_ppm', 'Rb_ppm', 'SO3_pct',
            'Sr_ppm', 'TiO2_pct', 'V_ppm', 'Zn_ppm', 'Zr_ppm']
        boreholes = dict()
        for name, data in vogue_data.groupby("Hole_ID"):
            # Construct the borehole instance first
            bh = boreholes[name] = pysiss.borehole.Borehole(name)

            # Generate an IntervalDataSet corresponding to the values in the
            # spreadsheet
            data = data.dropna()
            from_depths = numpy.asarray(data['From'])
            to_depths = numpy.asarray(data['To'])
            dataset = bh.add_interval_dataset(
                'geochemistry', from_depths, to_depths)

            # Add all the geochemistry values as Properties
            for key in property_keys:
                datum = numpy.asarray(
                    data[key].convert_objects(convert_numeric=True))
                dataset.add_property(property_types[key], datum)



""" file: test_borehole_utilities.py
    author: Jess Robertson & Ben Caradoc-Davies
            CSIRO Earth Science and Resource Engineering
    date: Tuesday May 14, 2013

    description: unittests for pysiss/borehole/utilities.py
"""

import unittest
import numpy
from pysiss.utilities import mask_all_nans


class TestMaskNans(unittest.TestCase):

    """ Testing masking nans function
    """

    def test_fail(self):
        "Function should fail with different sized input"
        self.assertRaises(ValueError, mask_all_nans,
                          numpy.arange(10),
                          numpy.arange(11),
                          numpy.arange(10))

    def test_convert(self):
        "Function should convert floats to numpy arrays"
        mask_all_nans(range(10), range(10), range(10))

    def test_fail_non_numeric(self):
        "Function should fail with non-numeric input"
        self.assertRaises(ValueError, mask_all_nans,
                          "i'm a string",
                          range(10))
